import { task, promptTablesConflicts, promptColumnsConflicts } from 'cli/components-api';
import fs from 'fs';
import * as differ from '../../../../differ';
import prepareMigration from 'migrationPreparator';

import { prepareAddValuesToEnumJson, prepareAlterTableColumnsJson, prepareCreateEnumJson, prepareCreateIndexesJson, prepareCreateTableJson, prepareDropIndexesJson, prepareDropTableJson, prepareRenameTableJson } from '../../../../jsonStatements';
import { fromJson } from '../../../../sqlgenerator';

export interface TablesResultData<T> {
  created: T[],
  renamed: RenamedObject<T>[],
  deleted: T[],
}

export interface ColumnsResultData<T extends Named> {
  created: T[],
  renamed: RenamedObject<T>[],
  deleted: T[],
}

export interface Named {
  name: string
}

export interface RenamedObject<T> {
  old: T,
  new: T
}

const two = (num: number) => {
  return num.toString().padStart(2, '0')
}

const migrate = async () => {
  const timeoutE = (ms: number) => new Promise((resolve) => setTimeout(resolve, ms)).then(() => {
    throw new Error('Checking error handling');
  });

  const timeout = (ms: number) => new Promise((resolve) => setTimeout(resolve, ms));

  const migrationRootFolder = 'drizzle' // or from config/params
  const { prev, cur } = await task('preparing data schema json snapshot', async ({ setTitle }) => {
    return prepareMigration(migrationRootFolder, './data/v2')
  });

  // const j1 = JSON.parse(fs.readFileSync('./out/v1_1634546745488.json', 'utf8'));
  // const j2 = JSON.parse(fs.readFileSync('./out/v2_1634549601361.json', 'utf8'));

  const diff = await task('preparing schemas diff', async ({ setTitle }) => {
    return differ.differ(prev, cur);
  });

  const { created, renamed, deleted } = await promptTablesConflicts({
    newTables: diff.addedTables,
    missingTables: diff.deletedTables,
  });

  const jsonStatements: any[] = []

  const jsonCreateTables = created.map(it => {
    return prepareCreateTableJson(it)
  })

  const jsonCreateIndexesForCreatedTables = created.map(it => {
    return prepareCreateIndexesJson(it.name, it.indexes)
  }).flat()

  const jsonDropTables = deleted.map(it => {
    return prepareDropTableJson(it)
  })

  const jsonRenameTables = renamed.map(it => {
    return prepareRenameTableJson(it.old, it.new)
  })

  // const createNewTables = created.map(it => {
  //   return prepareCreateTable(it)
  // })
  // console.log(createNewTables.join('\n').trim())

  // const deleteTables = deleted.map(it => {
  //   return prepareDropTable(it)
  // })
  // console.log(deleteTables.join('\n').trim())

  const renamedWithAlternations = differ.diffForRenamed(renamed)
  const allAltered = diff.alteredTablesWithColumns.concat(renamedWithAlternations)

  const allAlteredResolved = []
  for (const table of allAltered) {
    const result = await promptColumnsConflicts(table);

    const { deleted, added, ...ini } = table

    allAlteredResolved.push({ ...ini, ...result })
  }

  // console.log(JSON.stringify(allAlteredResolved, null, 2))

  // const renameTables = renamed.map(it => {
  //   return prepareRenameTable(it.from, it.to)
  // })
  // console.log(renameTables.join('\n').trim())

  const jsonAlterTables = allAlteredResolved.map(it => {
    return prepareAlterTableColumnsJson(it)
  }).flat()

  const jsonCreateIndexesForAllAlteredTables = allAlteredResolved.map(it => {
    return prepareCreateIndexesJson(it.name, it.addedIndexes || {})
  }).flat()

  const jsonDropIndexesForAllAlteredTables = allAlteredResolved.map(it => {
    return prepareDropIndexesJson(it.name, it.deletedIndexes || {})
  }).flat()

  // // Enums:
  // // - создание енама ✅
  // // - переименование енама (пока не делаю)⏳
  // // - добавление вэлью к енаму ✅
  // // - ренейм вейлью у енама (пока не делаю, это надо запрашивать опять же через слай)⏳
  // // - удаление енама -> чекать не используется ли где-то енам и сначала ранить миграции и в самом конце удаление енама⏳
  // // - удаление вэлью из енама -> блок ❌
  // const enums = result.addedEnums.map(it => {
  //   return prepareCreateEnum(it.name, it.values)
  // })

  const createEnums = diff.addedEnums.map(it => {
    return prepareCreateEnumJson(it.name, it.values)
  })

  //todo: block enum rename, enum value rename and enun deletion for now
  const jsonAlterEnumsWithAddedValues = diff.alteredEnums.map(it => {
    return prepareAddValuesToEnumJson(it.name, it.addedValues)
  }).flat()

  jsonStatements.push(...createEnums)
  jsonStatements.push(...jsonAlterEnumsWithAddedValues)
  jsonStatements.push(...jsonCreateTables)
  jsonStatements.push(...jsonCreateIndexesForCreatedTables)
  jsonStatements.push(...jsonDropTables)
  jsonStatements.push(...jsonRenameTables)
  jsonStatements.push(...jsonAlterTables)
  jsonStatements.push(...jsonCreateIndexesForAllAlteredTables)
  jsonStatements.push(...jsonDropIndexesForAllAlteredTables)

  // console.log(JSON.stringify(jsonStatements, null, 2))
  // console.log(jsonStatements)

  const sqlStatements = fromJson(jsonStatements)
  console.log(sqlStatements.join('\n'))

  // todo: save results to a new migration folder
  const date = new Date()
  // const folderName = `${date.getFullYear()}-${two(date.getMonth())}-${two(date.getDate())} ${two(date.getHours())}:${two(date.getMinutes())}:${two(date.getSeconds())}`
  const folderName = date.toISOString()
  const migrationFolderPath = `./${migrationRootFolder}/${folderName}`
  fs.mkdirSync(migrationFolderPath)
  fs.writeFileSync(`${migrationFolderPath}/snapshot.json`, JSON.stringify(cur, null, 2))
  fs.writeFileSync(`${migrationFolderPath}/migration.sql`, sqlStatements.join('\n'))

};

export default migrate;
