import React from 'react';
import { Box, Text } from 'ink';
import SelectInput from 'ink-select-input';
import Table, { Header } from 'ink-table';
import { useMachine } from '@xstate/react';
import { Item } from 'ink-select-input/build/SelectInput';

import { ColumnsResultData, Named } from 'cli/commands/migration/migrate';
import formatDataForTable from 'cli/utils/formatDataForTable';
import {
  Action,
  Confirmation,
  actions,
  confirmations,
} from 'cli/utils/valuesForPrompts';
import createResolveColumnsMachine, { TableWithColumn } from 'cli/machines/resolveColumnsMachine';

interface Props {
  onDone: (value: ColumnsResultData<Named>) => void
  tableWithColumns: TableWithColumn,
}

const PromptColumnsConflicts: React.FC<Props> = ({
  onDone,
  tableWithColumns: props,
}) => {
  const [current, send] = useMachine(
    createResolveColumnsMachine({ tableWithColumns: props }),
  );

  const {
    tableName,
    addedColumns,
    deletedColumns,
    missingItemIndex,
    created,
    renamed,
    deleted,
  } = current.context;

  const selectItemIndex = (item: Item<number>) => {
    send({ type: 'CHOICE_ITEM', itemIndex: item.value });
  };

  const selectNewItemIndex = (item: Item<number>) => {
    send({ type: 'CHOICE_NEW_ITEM', itemIndex: item.value });
  };

  const selectAction = (action: Item<Action>) => {
    send({ type: action.value });
  };

  const confirm = (confirmationItem: Item<Confirmation>) => {
    send({ type: confirmationItem.value });
  };

  const renderDOM = () => {
    if (current.matches('table')) {
      return (
        <>
          <Box display="flex" flexDirection="column">
            <Box display="flex">
              <Box flexDirection="column">
                <Header>{`Table "${tableName}" missing columns:`}</Header>
                <SelectInput
                  items={deletedColumns!.map((column, i: number) => ({
                    key: String(i),
                    label: column.name,
                    value: i,
                  }))}
                  onSelect={selectItemIndex}
                />
              </Box>
              {!!addedColumns?.length && (
                <Box flexDirection="column" paddingLeft={5}>
                  <Header>New tables:</Header>
                  <Box display="flex" flexDirection="column">
                    {addedColumns.map(({ name }, i) => <Text key={name + i}>{name}</Text>)}
                  </Box>
                </Box>
              )}
            </Box>
          </Box>
        </>
      );
    }

    if (current.matches('action.actionChoice')) {
      const tempActions = !addedColumns!.length
        ? actions.filter(({ value }) => value !== Action.RENAME)
        : actions;

      return (
        <>
          <Header>{`${deletedColumns![missingItemIndex]?.name} is:`}</Header>
          <SelectInput items={tempActions} onSelect={selectAction} />
        </>
      );
    }

    if (current.matches('confirmationDelete')) {
      return (
        <>
          <Header>!!! Data in table will be lost !!!</Header>
          <Text>Are you sure?</Text>
          <SelectInput items={confirmations} onSelect={confirm} />
        </>
      );
    }

    if (current.matches('confirmationRename')) {
      return (
        <>
          <Header>Are you sure?</Header>
          <SelectInput items={confirmations} onSelect={confirm} />
        </>
      );
    }

    if (current.matches('action.rename')) {
      return (
        <>
          <Header>{`${deletedColumns![missingItemIndex]?.name} was renamed to:`}</Header>
          <SelectInput
            items={addedColumns!.map((table, i: number) => ({
              key: String(i),
              label: table.name,
              value: i,
            }))}
            onSelect={selectNewItemIndex}
          />
        </>
      );
    }

    if (current.matches('done')) {
      onDone({ created, renamed, deleted });
    }

    return (<></>);
  };

  return (!!deleted.length
      || !!renamed.length
      || !!created.length) ? (
        <Box flexDirection="column" margin={1}>
          <Box flexDirection="column" marginBottom={0}>
            <Header>{`${tableName} columns:`}</Header>
            <Table
              data={formatDataForTable([
                {
                  title: 'Deleted',
                  values: deleted,
                },
                {
                  title: 'Renamed',
                  values: renamed,
                },
                {
                  title: 'Created',
                  values: created,
                },
              ])}
            />
          </Box>
          {renderDOM()}
        </Box>
    ) : renderDOM();
};

export default PromptColumnsConflicts;
