import React from 'react';
import { Box, Text } from 'ink';
import SelectInput from 'ink-select-input';
import Table, { Header } from 'ink-table';
import { useMachine } from '@xstate/react';
import { Item } from 'ink-select-input/build/SelectInput';

import { Named, TablesResultData } from 'cli/commands/migration/migrate';
import createResolveTablesMachine from 'cli/machines/resolveTablesMachine';
import formatDataForTable from 'cli/utils/formatDataForTable';
import {
  Action,
  Confirmation,
  actions,
  confirmations,
} from 'cli/utils/valuesForPrompts';

export interface RenderTablesSelectProps<T extends Named> {
  missingTables: T[],
  newTables: T[],
}

interface Props<T extends Named> extends RenderTablesSelectProps<T> {
  onDone: (value: TablesResultData<T>) => void;
}

const PromptTablesConflicts: React.FC<Props<any>> = ({
  onDone,
  newTables: newProps,
  missingTables: missingProps,
}) => {
  const [current, send] = useMachine(createResolveTablesMachine({
    missingTables: missingProps,
    newTables: newProps,
  }));

  const {
    missingItemIndex,
    newTables,
    missingTables,
    createdTables,
    renamedTables,
    deletedTables,
  } = current.context;

  const selectItemIndex = (item: Item<number>) => {
    send({ type: 'CHOICE_ITEM', itemIndex: item.value });
  };

  const selectNewItemIndex = (item: Item<number>) => {
    send({ type: 'CHOICE_NEW_ITEM', itemIndex: item.value });
  };

  const selectAction = (action: Item<Action>) => {
    send({ type: action.value });
  };

  const confirm = (confirmationItem: Item<Confirmation>) => {
    send({ type: confirmationItem.value });
  };

  const renderDOM = () => {
    if (current.matches('table')) {
      return (
        <>
          <Box display="flex" flexDirection="column">
            <Box display="flex">
              <Box flexDirection="column">
                <Header>Missing tables:</Header>
                <SelectInput
                  items={missingTables!.map((table, i: number) => ({
                    key: String(i),
                    label: table.name,
                    value: i,
                  }))}
                  onSelect={selectItemIndex}
                />
              </Box>
              {!!newTables?.length && (
                <Box flexDirection="column" paddingLeft={5}>
                  <Header>New tables:</Header>
                  <Box display="flex" flexDirection="column">
                    {newTables.map(({ name }, i) => <Text key={name + i}>{name}</Text>)}
                  </Box>
                </Box>
              )}
            </Box>
          </Box>
        </>
      );
    }

    if (current.matches('action.actionChoice')) {
      const tempActions = !newTables!.length
        ? actions.filter(({ value }) => value !== Action.RENAME)
        : actions;

      return (
        <>
          <Header>{`${missingTables![missingItemIndex]?.name} is:`}</Header>
          <SelectInput items={tempActions} onSelect={selectAction} />
        </>
      );
    }

    if (current.matches('confirmationDelete')) {
      return (
        <>
          <Header>!!! Data in table will be lost !!!</Header>
          <Text>Are you sure?</Text>
          <SelectInput items={confirmations} onSelect={confirm} />
        </>
      );
    }

    if (current.matches('confirmationRename')) {
      return (
        <>
          <Header>Are you sure?</Header>
          <SelectInput items={confirmations} onSelect={confirm} />
        </>
      );
    }

    if (current.matches('action.rename')) {
      return (
        <>
          <Header>{`${missingTables![missingItemIndex]?.name} was renamed to:`}</Header>
          <SelectInput
            items={newTables!.map((table, i: number) => ({
              key: String(i),
              label: table.name,
              value: i,
            }))}
            onSelect={selectNewItemIndex}
          />
        </>
      );
    }

    if (current.matches('done')) {
      onDone({
        created: createdTables,
        renamed: renamedTables,
        deleted: deletedTables,
      });
    }

    return (<></>);
  };

  return (!!deletedTables.length
      || !!renamedTables.length
      || !!createdTables.length) ? (
        <Box flexDirection="column" margin={1}>
          <Box flexDirection="column" marginBottom={0}>
            <Header>Tables: </Header>
            <Table
              data={formatDataForTable([
                {
                  title: 'Deleted',
                  values: deletedTables,
                },
                {
                  title: 'Renamed',
                  values: renamedTables,
                },
                {
                  title: 'Created',
                  values: createdTables,
                },
              ])}
            />
          </Box>
          {renderDOM()}
        </Box>
    ) : renderDOM();
};

export default PromptTablesConflicts;
