import { assign, createMachine, send } from 'xstate';
import { Named, RenamedObject } from 'cli/commands/migration/migrate';

type Event =
  | { type: 'CHOICE_ITEM', itemIndex: number }
  | { type: 'DELETE' }
  | { type: 'RENAME' }
  | { type: 'CANCEL' }
  | { type: 'CONFIRM' }
  | { type: 'NEXT' }
  | { type: 'CHOICE_NEW_ITEM'; itemIndex: number };

export interface TableWithColumn {
  name: string;
  added: Named[],
  deleted: Named[],
}

interface Props {
  tableWithColumns: TableWithColumn,
}

interface Context {
  tableName: string,
  addedColumns: Named[],
  deletedColumns: Named[],
  missingItemIndex: number,
  newItemIndex: number,
  created: Named[];
  renamed: RenamedObject<Named>[];
  deleted: Named[];
}

const createResolveColumnsMachine = (props: Props) => (
  createMachine<Context, Event>({
    id: 'resolveColumns',
    initial: 'table',
    context: {
      tableName: props.tableWithColumns.name,
      addedColumns: props.tableWithColumns.added,
      deletedColumns: props.tableWithColumns.deleted,
      missingItemIndex: 0,
      newItemIndex: 0,
      created: [],
      renamed: [],
      deleted: [],
    },
    states: {
      table: {
        entry: send({ type: 'NEXT' }),
        on: {
          NEXT: [
            {
              target: 'done',
              cond: 'isMissingColumnsResolved',
              actions: ['resolveRemaining'],
            },
            {
              target: 'done',
              cond: 'isNewColumnsResolved',
              actions: ['resolveMissing'],
            },
          ],
          CHOICE_ITEM: { target: 'action', actions: ['choseItem'] },
        },
      },
      action: {
        initial: 'actionChoice',
        states: {
          actionChoice: {
            on: { DELETE: '#resolveColumns.confirmationDelete', RENAME: 'rename' },
          },
          rename: {
            on: {
              CHOICE_NEW_ITEM: { target: '#resolveColumns.confirmationRename', actions: ['choseNewItem'] },
            },
          },
        },
      },
      confirmationDelete: {
        on: {
          CANCEL: 'action.actionChoice',
          CONFIRM: [
            { target: 'final', actions: ['delete'] },
          ],
        },
      },
      confirmationRename: {
        on: {
          CANCEL: 'action.actionChoice',
          CONFIRM: [
            { target: 'final', actions: ['rename'] },
          ],
        },
      },
      final: {
        entry: send({ type: 'NEXT' }),
        on: {
          NEXT: [
            {
              target: 'done',
              cond: 'isMissingColumnsResolved',
              actions: ['resolveRemaining'],
            },
            {
              target: 'table',
            },
          ],
        },
      },
      done: {},
    },
  },
  {
    guards: {
      isMissingColumnsResolved: ({ deletedColumns }) => !deletedColumns.length,
      isNewColumnsResolved: ({ addedColumns }) => !addedColumns.length,
    },
    actions: {
      choseItem: assign({
        missingItemIndex: (context, event) => (event.type === 'CHOICE_ITEM' ? event.itemIndex : 0),
      }),

      choseNewItem: assign({
        newItemIndex: (context, event) => (event.type === 'CHOICE_NEW_ITEM' ? event.itemIndex : 0),
      }),

      delete: assign({
        deleted: ({
          missingItemIndex,
          deleted,
          deletedColumns,
        }) => [...deleted, deletedColumns[missingItemIndex]],
        deletedColumns: ({
          missingItemIndex,
          deletedColumns,
        }) => deletedColumns.filter((_, index) => index !== missingItemIndex),
      }),

      rename: assign({
        renamed: ({
          missingItemIndex,
          newItemIndex,
          renamed,
          addedColumns,
          deletedColumns,
        }) => [
          ...renamed,
          { old: deletedColumns[missingItemIndex], new: addedColumns[newItemIndex] },
        ],
        deletedColumns: ({
          missingItemIndex,
          deletedColumns,
        }) => deletedColumns.filter((_, index) => index !== missingItemIndex),
        addedColumns: ({
          newItemIndex,
          addedColumns,
        }) => addedColumns.filter((_, index) => index !== newItemIndex),
      }),

      resolveRemaining: assign({
        created: ({
          addedColumns,
          created,
        }) => [...created, ...addedColumns],
        addedColumns: (context) => [],
      }),

      resolveMissing: assign({
        deleted: ({
          deletedColumns,
          deleted,
        }) => [...deleted, ...deletedColumns],
        deletedColumns: (context) => [],
      }),
    },
  }));

export default createResolveColumnsMachine;
