import { assign, createMachine, send } from 'xstate';
import { Named, RenamedObject } from 'cli/commands/migration/migrate';

type Event =
  | { type: 'CHOICE_ITEM', itemIndex: number }
  | { type: 'DELETE' }
  | { type: 'RENAME' }
  | { type: 'CANCEL' }
  | { type: 'CONFIRM' }
  | { type: 'NEXT' }
  | { type: 'CHOICE_NEW_ITEM'; itemIndex: number };

interface Context extends Props {
  missingItemIndex: number,
  newItemIndex: number,
  createdTables: Named[];
  renamedTables: RenamedObject<Named>[];
  deletedTables: Named[];
}

interface Props {
  missingTables: Named[],
  newTables: Named[],
}

const createResolveTablesMachine = (props: Props) => (
  createMachine<Context, Event>({
    id: 'resolveTables',
    initial: 'table',
    context: {
      ...props,
      missingItemIndex: 0,
      newItemIndex: 0,
      createdTables: [],
      renamedTables: [],
      deletedTables: [],
    },
    states: {
      table: {
        entry: send({ type: 'NEXT' }),
        on: {
          NEXT: [
            {
              target: 'done',
              cond: 'isMissingTablesResolved',
              actions: ['resolveRemaining'],
            },
            {
              target: 'done',
              cond: 'isNewTablesResolved',
              actions: ['resolveMissing'],
            },
          ],
          CHOICE_ITEM: { target: 'action', actions: ['choseItem'] },
        },
      },
      action: {
        initial: 'actionChoice',
        states: {
          actionChoice: {
            on: { DELETE: '#resolveTables.confirmationDelete', RENAME: 'rename' },
          },
          rename: {
            on: {
              CHOICE_NEW_ITEM: { target: '#resolveTables.confirmationRename', actions: ['choseNewItem'] },
            },
          },
        },
      },
      confirmationDelete: {
        on: {
          CANCEL: 'action.actionChoice',
          CONFIRM: [
            { target: 'final', actions: ['delete'] },
          ],
        },
      },
      confirmationRename: {
        on: {
          CANCEL: 'action.actionChoice',
          CONFIRM: [
            { target: 'final', actions: ['rename'] },
          ],
        },
      },
      final: {
        entry: send({ type: 'NEXT' }),
        on: {
          NEXT: [
            {
              target: 'done',
              cond: 'isMissingTablesResolved',
              actions: ['resolveRemaining'],
            },
            {
              target: 'table',
            },
          ],
        },
      },
      done: {},
    },
  },
  {
    guards: {
      isMissingTablesResolved: ({ missingTables }) => !missingTables.length,
      isNewTablesResolved: ({ newTables }) => !newTables.length,
    },
    actions: {
      choseItem: assign({
        missingItemIndex: (context, event) => (event.type === 'CHOICE_ITEM' ? event.itemIndex : 0),
      }),

      choseNewItem: assign({
        newItemIndex: (context, event) => (event.type === 'CHOICE_NEW_ITEM' ? event.itemIndex : 0),
      }),

      delete: assign({
        deletedTables: ({
          missingItemIndex,
          deletedTables,
          missingTables,
        }) => [...deletedTables, missingTables[missingItemIndex]],
        missingTables: ({
          missingItemIndex,
          missingTables,
        }) => missingTables.filter((_, index) => index !== missingItemIndex),
      }),

      rename: assign({
        renamedTables: ({
          missingItemIndex,
          newItemIndex,
          renamedTables,
          newTables,
          missingTables,
        }) => [
          ...renamedTables,
          { old: missingTables[missingItemIndex], new: newTables[newItemIndex] },
        ],
        missingTables: ({
          missingItemIndex,
          missingTables,
        }) => missingTables.filter((_, index) => index !== missingItemIndex),
        newTables: ({
          newItemIndex,
          newTables,
        }) => newTables.filter((_, index) => index !== newItemIndex),
      }),

      resolveRemaining: assign({
        createdTables: ({
          newTables,
          createdTables,
        }) => [...createdTables, ...newTables],
        newTables: (context) => [],
      }),

      resolveMissing: assign({
        deletedTables: ({
          missingTables,
          deletedTables,
        }) => [...deletedTables, ...missingTables],
        missingTables: (context) => [],
      }),
    },
  }));

export default createResolveTablesMachine;
