export class Types {
    static createTable = 'create_table'
    static dropTable = 'drop_table'
    static renameTable = 'rename_table'
    static alterTableDropColumn = 'alter_table_drop_column'
    static alterTableAddColumn = 'alter_table_add_column'
    static alterTableAlterColumnSetType = 'alter_table_alter_column_set_type'
    static alterTableAlterColumnSetDefault = 'alter_table_alter_column_set_default'
    static alterTableAlterColumnDropDefault = 'alter_table_alter_column_drop_default'
    static alterTableAlterColumnSetNotNull = 'alter_table_alter_column_set_notnull'
    static alterTableAlterColumnDropNotNull = 'alter_table_alter_column_drop_notnull'
    static createIndex = 'create_index'
    static dropIndex = 'drop_index'
    static createTypeEnum = 'create_type_enum'
    static alterTypeAddValue = 'alter_type_add_value'
}

export const prepareCreateTableJson = (table) => {
    const { name, columns } = table
    const mappedColumns = Object.keys(columns)
        .map(it => {
            return { ...columns[it] }
        })

    return {
        type: Types.createTable,
        tableName: name,
        columns: mappedColumns
    };
}

export const prepareDropTableJson = (table) => {
    return {
        type: Types.dropTable,
        tableName: table.name
    }
}

export const prepareRenameTableJson = (tableFrom, tableTo) => {
    return {
        type: Types.renameTable,
        from: tableFrom.name,
        to: tableTo.name
    }
}

export const prepareCreateEnumJson = (name, values) => {
    return {
        type: Types.createTypeEnum,
        enumName: name,
        values
    }
}

// https://blog.yo1.dog/updating-enum-values-in-postgresql-the-safe-and-easy-way/
export const prepareAddValuesToEnumJson = (name, values) => {
    return values.map(it => {
        return {
            type: Types.alterTypeAddValue,
            enumName: name,
            enumValue: it
        }
    })
}

export const prepareAlterTableColumnsJson = (table) => {
    const tableName = table['name'];
    const statements = []

    const dropColumns = _prepareDropColumns(tableName, table.deleted)
    const addColumns = _prepareAddColumns(tableName, table.added)
    const alterColumns = _prepareAlterColumns(tableName, table.altered)


    statements.push(...dropColumns)
    statements.push(...addColumns)
    statements.push(...alterColumns)

    return statements;
}

const _prepareDropColumns = (taleName, columns) => {
    return columns.map(it => {
        return {
            type: Types.alterTableDropColumn,
            tableName: taleName,
            columnName: it.name
        }
    })
}

const _prepareAddColumns = (tableName, columns) => {
    
    return columns.map(it => {
        return {
            type: Types.alterTableAddColumn,
            tableName,
            column: it
        }
    })
}

const _prepareAlterColumns = (tableName, columns) => {
    let statements = []

    for (const column of columns) {
        const columnName = column.name;

        if (column['type']) {
            const { __new } = column.type

            statements.push({
                type: Types.alterTableAlterColumnSetType,
                tableName,
                columnName,
                newDataType: __new
            })
        }

        if (column['defaultValue'] || column['defaultValue__added']) {
            const newDefaultValue = column.defaultValue.__new | column.defaultValue__added
            statements.push({
                type: Types.alterTableAlterColumnSetDefault,
                tableName,
                columnName,
                newDefault: newDefaultValue
            })
        }

        if (column['defaultValue__deleted']) {
            statements.push({
                type: Types.alterTableAlterColumnDropDefault,
                tableName,
                columnName,
            })
        }

        if (column['notNull__added'] || (column['notNull'] && column['notNull']['__new'])) {
            statements.push({
                type: Types.alterTableAlterColumnSetNotNull,
                tableName,
                columnName
            })
        }

        if (column['notNull__deleted'] || (column['notNull'] && !column['notNull']['__new'])) {
            statements.push({
                type: Types.alterTableAlterColumnDropNotNull,
                tableName,
                columnName
            })
        }
    }

    return statements
}

export const prepareCreateIndexesJson = (tableName, indexes) => {
    return indexes.map(index => {
        return {
            type: Types.createIndex,
            tableName,
            indexName: index.name,
            indexValue: index.columns.join(', ')
        }
    })
}
export const prepareDropIndexesJson = (tableName, indexes) => {
    return indexes.map(index => {
        return {
            type: Types.dropIndex,
            tableName,
            indexName: index.name,
        }
    })
}
